/*
AgeUnivers - algorithme de simulation pour SimuApplet

Copyright (C) 2002 Observatoire de Paris-Meudon

Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier conformment aux dispositions de la Licence Publique Gnrale GNU, telle que publie par la Free Software Foundation ; version 2 de la licence, ou encore ( votre choix) toute version ultrieure.

Ce programme est distribu dans l'espoir qu'il sera utile, mais SANS AUCUNE GARANTIE ; sans mme la garantie implicite de COMMERCIALISATION ou D'ADAPTATION A UN OBJET PARTICULIER. Pour plus de dtail, voir la Licence Publique Gnrale GNU .

Vous devez avoir reu un exemplaire de la Licence Publique Gnrale GNU en mme temps que ce programme ; si ce n'est pas le cas, crivez  la Free Software Foundation Inc., 675 Mass Ave, Cambridge, MA 02139, Etats-Unis.

*/

import simu.*;

/**
 * Algorithme de simulation pour SimuApplet.
 * Ce programme calcule l'age de l'univers et l'volution du facteur d'echelle "a" en fonction du temps "t"
 * Les modeles d'Univers consideres sont ceux de Friedmann-Lemaitre-Robertson-Walker (AgeUnivers)
 */
public class AgeUnivers extends SimuApplet {

	public static final double mpc=3.08567758074e22;
	public static final double gyr=3.1688087814e-17,g=6.673231e-11;
	public static final double c=2.99792458e8;
	public static final double twopi=6.283185307179586;
	public static final double a0 = 1;
	double eres[];
	String eres2[];
	double h0,Omega_m0,Omega_rad0,Omega_l0,Omega_k0,T0,temp;
	double rho_m,rho_rad,rho_lam;
	double stop = 1e-5;
	double da;
	double a,t,h,integral,age,courbure;
	double tmp;
	boolean nobb,nobc,failed,half; 
	
	// constructeur de la classe, avec la dfinition des paramtres de simulation
	// et des options d'affichage.
    
	public AgeUnivers() {
        
	// paramtres d'entre
        ParamIn h0 = new ParamIn();
        h0.type = "nombre";
        h0.label = "h0";
        h0.titre = "Constante de Hubble aujourd'hui";
        h0.unit = "km/s/Mpc";
        h0.acquisition = "champ";
        h0.defaut = "70";
        addParamIn(h0);

        ParamIn Omega_m0 = new ParamIn();
        Omega_m0.type = "nombre";
        Omega_m0.label = "Omega_m0";
        Omega_m0.titre = "Densit de matire aujourd'hui";
        Omega_m0.unit = "densit critique";
        Omega_m0.acquisition = "champ";
        Omega_m0.defaut = "0.3";
        addParamIn(Omega_m0);

	ParamIn Omega_l0 = new ParamIn();
        Omega_l0.type = "nombre";
        Omega_l0.label = "Omega_l0";
        Omega_l0.titre = "Densit d'nergie noire aujourd'hui ";
        Omega_l0.unit = "densit critique";
        Omega_l0.acquisition = "champ";
        Omega_l0.defaut = "0.7";
        addParamIn(Omega_l0);

	ParamIn Omega_rad0 = new ParamIn();
        Omega_rad0.type = "nombre";
        Omega_rad0.label = "Omega_rad0";
        Omega_rad0.titre = "Temprature du rayonnement fossile ";
        Omega_rad0.unit = "K";
        Omega_rad0.acquisition = "champ";
        Omega_rad0.defaut = "2.75";
        addParamIn(Omega_rad0);


	// paramtres de sortie
    ParamOut t = new ParamOut();
    t.type = "nombre";
    t.label = "t";
	t.unit = "Gyr";
	t.titre = "Temps (Milliards d'annes)";
    addParamOut(t);
    
	ParamOut a = new ParamOut();
    a.type = "nombre";
    a.label = "a";
	a.titre = "a";
    addParamOut(a);

	ParamOut rho_m = new ParamOut();
    rho_m.type = "nombre";
    rho_m.label = "rho_m";
	rho_m.titre = "rho";
    addParamOut(rho_m);

	ParamOut rho_rad = new ParamOut();
    rho_rad.type = "nombre";
    rho_rad.label = "rho_rad";
	rho_rad.titre = "rho_rad";
    addParamOut(rho_rad);

	ParamOut rho_lam = new ParamOut();
    rho_lam.type = "nombre";
    rho_lam.label = "rho_lam";
	rho_lam.titre = "rho_lam";
    addParamOut(rho_lam);

    ParamOut temp = new ParamOut();
    temp.type = "nombre";
    temp.label = "temp";
	temp.titre = "Temperature";
    addParamOut(temp);

	ParamOut comment = new ParamOut();
	comment.type = "string";
	comment.label = "comment";
	addParamOut(comment);

	
	// options d'affichage
	// affichage "multiplot"
        Affichage multi = new Affichage();
        multi.type = "multiplot";
        multi.titre = "Evolution cosmologique";
        multi.sousAff = new Affichage[2];

	// affichage "plotlive"
        Affichage p = new Affichage();
        p.type = "plotlive";
        p.titre = "Facteur d'chelle a(t)";
		String[][] pparams = {{"t", "a"}};
        p.params = pparams;
//        p.fond = "hudf.gif";
		//p.ajuster=true;
		p.carre=true;
		p.connecter=true;
		p.points="none";
		p.xdebut = ""+-15;
		p.xfin = ""+1;
		p.ydebut = ""+0;
		p.yfin = ""+1;
		multi.sousAff[0] = p;

		// affichage "plotlive"
        Affichage p2 = new Affichage();
        p2.type = "plotlive";
        p2.titre = "Densits rho(t)";
	String[][] pparams2 = {{"a", "rho_m"},{"a", "rho_rad"},{"a", "rho_lam"}};
        p2.params = pparams2;
//        p2.fond = "hudf.gif";
	p2.ajuster=true;
	p2.carre=true;
	p2.connecter=true;
	p2.points="none";
	p2.ylog=true;
	p2.xlog=true;
	multi.sousAff[1] = p2;
 
        
	addAffichage(multi);

	// affichage "tableau"
        Affichage tab = new Affichage();
        tab.type = "tableau";
        tab.titre = "Eres cosmologiques";
		String[][] params_tab = {{"t", "temp","comment"}};
        tab.params = params_tab;
        addAffichage(tab);

	}

	// calcul live (utilis avec affichage "plotlive")
    // initialisation
    public void initCalculLive(ListeValeurs in) throws SimuException {
       h0 = in.lireDouble("h0");
    	Omega_m0 = in.lireDouble("Omega_m0");
	Omega_l0 = in.lireDouble("Omega_l0");
	Omega_rad0 = in.lireDouble("Omega_rad0");
	Omega_rad0=4.1534e-5/h0/h0*100*100*Math.pow((Omega_rad0/2.725),4);

	a = a0;
	t = 0;
	da=1e-2;

	h0=h0*1.e3/mpc;
	Omega_k0=Omega_rad0+Omega_m0+Omega_l0-1.;
	
	if (Omega_m0<=0.5){
	tmp=(1-Omega_m0)/Omega_m0;
	tmp=Math.log(tmp+Math.sqrt(tmp*tmp-1.))/3;
	tmp=(Math.exp(tmp)+Math.exp(-tmp))/2;
	tmp=4*Omega_m0*Math.pow(tmp,3.);
	}

	if (Omega_m0>0.5){
	tmp=(1-Omega_m0)/Omega_m0;
	tmp=4*Omega_m0*Math.pow(Math.cos(Math.acos(tmp)/3),3.);
	}
	
	nobb=false;
	if(Omega_l0>=tmp){
	nobb=true;
	}

	if(Omega_m0>1){
	tmp=(1-Omega_m0)/Omega_m0;
	tmp=4*Omega_m0*Math.pow(Math.cos(Math.acos(tmp)/3)+4/3*Math.PI,3.);
	}
	if(Omega_m0<=1){
	tmp=0;
	}
	nobc=true;
	if(Omega_l0<=tmp){
	nobc=false;
	}

	
	if(!nobc){
		a=1e-3;
		half=false;
	}

    }

	// Calcul de a  a+da
    public ListeValeurs calculLive() throws SimuException {
	
	ListeValeurs out = new ListeValeurs();

	if(nobc){
	if( (!nobb & a<1e-4))
		return null; // fin du calcul
	
	if(nobb & failed )
		return null; // fin du calcul
	
	if(a<=1e-2)
		da=1e-4;
    
	rombint(a,a-da,stop);
	t +=integral;
	a += -da;

	rho_m=Omega_m0*3.*h0*h0/4./twopi/g*Math.pow((a0/a),3.);
	rho_rad=Omega_rad0*3.*h0*h0/4./twopi/g*Math.pow((a0/a),4.);
	rho_lam=Omega_l0*3.*h0*h0/4./twopi/g;

	out.ajouterDouble("t", t);
	out.ajouterDouble("a", a);
	out.ajouterDouble("temp", T0/a);
	out.ajouterDouble("rho_m", rho_m);
	out.ajouterDouble("rho_rad", rho_rad);
	out.ajouterDouble("rho_lam", rho_lam);
	age=Math.floor(10.*t)/10.;
	age=-age;
	//courbure=Math.floor(100.*Omega_k0)/10.;
	if(Omega_k0>0){
		out.ajouterString("comment","Univers elliptique ; Age de l'Univers : "+age+" Milliards d'annes ");
	}
	if(Omega_k0<0){
		out.ajouterString("comment","Univers hyperbolique ; Age de l'Univers : "+age+" Milliards d'annes ");
	}
	if(Omega_k0==0){
		out.ajouterString("comment","Univers plat ; Age de l'Univers : "+age+" Milliards d'annes ");
	}
	return out;
	}
	
	if(!nobc){
	tmp=Omega_m0 * Math.pow(a0/a,3.)+Omega_rad0 * Math.pow(a0/a,4.)-Omega_k0 * Math.pow(a0/a,2.)+Omega_l0;
	
//	if( tmp<=1e-5 & half)
	if( failed & half)
		return null; // fin du calcul
	
	if(a<=1e-2)
		da=1e-4;

    	if(a>1e-2)
		da=1e-2;
	rombint(a,a+da,stop);
	if(failed & !half) {
		half=true;
		a+=-1e-2;
		rombint(a,a+da,stop);
		//System.err.println("tmp "+tmp+" nobc "+nobc);
	}
	if (!failed) 
		t +=integral;

	if(!half)
		a += da;
	if(half)
		a+= -da;

	rho_m=Omega_m0*3.*h0*h0/4./twopi/g*Math.pow((a0/a),3.);
	rho_rad=Omega_rad0*3.*h0*h0/4./twopi/g*Math.pow((a0/a),4.);
	rho_lam=Omega_l0*3.*h0*h0/4./twopi/g;

	out.ajouterDouble("t", t);
	out.ajouterDouble("a", a);
	out.ajouterDouble("temp", T0/a);
	out.ajouterDouble("rho_m", rho_m);
	out.ajouterDouble("rho_rad", rho_rad);
	out.ajouterDouble("rho_lam", rho_lam);
	age=Math.floor(10.*t)/10.;
	age=age;
	if(Omega_k0>0){
		out.ajouterString("comment","Univers elliptique ; Age de l'Univers : "+age+" Milliards d'annes ");
	}
	if(Omega_k0<0){
		out.ajouterString("comment","Univers hyperbolique ; Age de l'Univers : "+age+" Milliards d'annes ");
	}
	if(Omega_k0==0){
		out.ajouterString("comment","Univers plat ; Age de l'Univers : "+age+" Milliards d'annes ");
	}
	return out;
	}
	return null;
	}

	// calcul complet d'un coup (utilis avec les affichages "plot" et "tableau")
	public ListeValeurs calcul(ListeValeurs in) throws SimuException {
    
	ListeValeurs out = new ListeValeurs();
	initParamIn(in);      
	
	da=1e-3;
	while (a>=1/eres[1]) {
	
	rombint(a,a-da,stop);
	t +=integral;

	a += -da;

    }
	age=Math.abs(t);
	out.ajouterDouble("t", age/gyr);
    	out.ajouterDouble("temp", eres[0]*T0);
	out.ajouterString("comment","Aujourd'hui");
	

	t=Math.pow(1./eres[1]/Math.pow(2.*h0*h0*Omega_rad0,0.25)/a0,2.);

	out.ajouterDouble("t", t);
	out.ajouterDouble("temp", eres[1]);
	out.ajouterString("comment",eres2[0]);
	
	t=Math.pow(1./eres[2]/Math.pow(2.*h0*h0*Omega_rad0,0.25)/a0,2.);

	out.ajouterDouble("t", t);
	out.ajouterDouble("temp", eres[2]);
	out.ajouterString("comment",eres2[1]);
	
	t=Math.pow(1./eres[3]/Math.pow(2.*h0*h0*Omega_rad0,0.25)/a0,2.);

	out.ajouterDouble("t", t);
	out.ajouterDouble("temp", eres[3]);
	out.ajouterString("comment",eres2[2]);
	
	t=Math.pow(1./eres[4]/Math.pow(2.*h0*h0*Omega_rad0,0.25)/a0,2.);

	out.ajouterDouble("t", t);
	out.ajouterDouble("temp", eres[4]);
	out.ajouterString("comment",eres2[3]);
	
	out.ajouterDouble("t", 5.391e-44);
    	out.ajouterDouble("temp", eres[5]);
	out.ajouterString("comment",eres2[4]);
	
	out.ajouterDouble("t", 0);
	out.ajouterString("comment",eres2[5]);

	return out;
	}
	//initialisation des paramtres d'entre
    public void initParamIn(ListeValeurs in) throws SimuException{
    
	h0 = in.lireDouble("h0");
	Omega_m0 = in.lireDouble("Omega_m0");
	Omega_l0 = in.lireDouble("Omega_l0");
	Omega_rad0 = in.lireDouble("Omega_rad0");
	T0=Omega_rad0;
	Omega_rad0=4.1534e-5/h0/h0*100*100*Math.pow((Omega_rad0/2.725),4);

	a = a0;
	t = 0;

	h0=h0*1.e3/mpc;
	Omega_k0=Omega_rad0+Omega_m0+Omega_l0-1.;
    
	eres = new double[6];
	eres[0]=1; //Aujourd'hui
	eres[1]=1e3; // CMB
	eres[2]=1e10; // Ere Leptonique
	eres[3]=1e12; // Ere Hadronique
	eres[4]=1e28; // Grande Unification
	eres[5]=1e32; // Temps de Planck

	eres2 = new String[6];
	eres2[0]="Ere Stellaire";
	eres2[1]="Ere Radiative";
	eres2[2]="Ere Leptonique";
	eres2[3]="Ere Hadronique";
	eres2[4]="Grande Unification";
	eres2[5]="Ere de Planck";

    }
    
	//calcul du taux d'expansion en fonction de a
    public double hubble(double scale) {
	h=Omega_m0 * Math.pow(a0/scale,3.)+Omega_rad0 * Math.pow(a0/scale,4.)-Omega_k0 * Math.pow(a0/scale,2.)+Omega_l0;
	h=h0 * Math.sqrt(h);
	return h;
    }

    public void rombint(double x, double y, double tol) {
	//  Rombint returns the integral from a to b of f(x)dx using Romberg 
	// integration.
	//  The method converges provided that f(x) is continuous in (a,b).  
	//  The function f must be double precision and must be declared external 
	//  in the calling routine.  
	//  tol indicates the desired relative accuracy in the integral.

	double step,gmax,error,g0,fourj,g1;
	int nint,i,jmax;
	int maxiter=20,maxj=5;
	double g[];
	g= new double[maxj+2];
	
	failed=false;

	step=0.5*(y-x);
	gmax=step/y/hubble(y)*gyr;
	gmax+=step/x/hubble(x)*gyr;
	g[1]=gmax;
	nint=1;
	error=1.e20;
	i=0;
	g0=0;
	jmax=0;
	while (!(i>maxiter|(i>5 & Math.abs(error)< tol)) ){
	i++;
	//	Calculate next trapezoidal rule approximation to integral.

	g0=0;
	for(int k=0; k<nint; k++){
	   g0=g0+1./hubble(x+(k+k-1)*step)/(x+(k+k-1)*step)*gyr;
	}
	g0=0.5*g[1]+step*g0;
	step=0.5*step;
	nint=nint+nint;
	jmax=Math.min(i,maxj);
	fourj=1;
	for(int j=0; j<jmax; j++){
	//	Use Richardson extrapolation.
	   fourj=4.*fourj;
	   g1=g0+(g0-g[j])/(fourj-1.);
	   g[j]=g0;
	   g0=g1;
	}
	if (Math.abs(g0)>tol) {
	   error=1.-gmax/g0;
	   }
	else{
	   error=gmax;
	}
	gmax=g0;
	g[jmax+1]=g0;
	}
	integral=g0;
	//integral=da/x/h*gyr;
	if (i>maxiter) { // & Math.abs(error)>tol) {
	System.err.println("Rombint failed to converge "+jmax);
	failed=true;
	}

	}
}

